/* for finding memory leaks in debug mode with Visual Studio */
#if defined _DEBUG && defined _MSC_VER
#include <crtdbg.h>
#endif


#include <stdint.h>
#ifndef _WIN32
#include <unistd.h> /* usleep() */
#endif
#include "ft2_header.h"
#include "ft2_config.h"
#include "ft2_audio.h"
#include "ft2_gui.h"
#include "ft2_midi.h"
#include "ft2_gfxdata.h"
#include "ft2_scopes.h"
#include "ft2_mouse.h"
#include "ft2_video.h"

#define SCOPE_DATA_HEIGHT 36

/* data to be read from main update thread during sample trigger */
typedef struct scopeState_t
{
    int8_t *pek;
    uint8_t typ;
    int32_t len, repS, repL, playOffset;
} scopeState_t;

/* actual scope data */
typedef struct scope_t
{
    const int8_t *data;
    int8_t volume;
    uint8_t latchFlag, wasCleared, sample16bit, loop;
    int32_t direction, drawDirection, loopStart, loopEnd, loopLen, len, readPos;
    double dDrawDelta;
    float fDelta, fFrac;
} scope_t;

static volatile uint8_t scopesUpdating, scopesDisplaying;
static uint64_t timeNext64;
static scope_t scope[MAX_VOICES];
static SDL_Thread *scopeThread;
static scopeState_t scopeNewState[MAX_VOICES];

static const uint16_t scopeLenTab[16][32] =
{
    /*  2 ch */ {285, 285},
    /*  4 ch */ {141,141,141,141},
    /*  6 ch */ {93,93,93,93,93,93},
    /*  8 ch */ {69,69,69,69,69,69,69,69},
    /* 10 ch */ {55,55,55,54,54,55,55,55,54,54},
    /* 12 ch */ {45,45,45,45,45,45,45,45,45,45,45,45},
    /* 14 ch */ {39,38,38,38,38,38,38,39,38,38,38,38,38,38},
    /* 16 ch */ {33,33,33,33,33,33,33,33,33,33,33,33,33,33,33,33},
    /* 18 ch */ {29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29},
    /* 20 ch */ {26,26,26,26,26,26,26,26,25,25,26,26,26,26,26,26,26,26,25,25},
    /* 22 ch */ {24,24,23,23,23,23,23,23,23,23,23,24,24,23,23,23,23,23,23,23,23,23},
    /* 24 ch */ {21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21},
    /* 26 ch */ {20,20,19,19,19,19,19,19,19,19,19,19,19,20,20,19,19,19,19,19,19,19,19,19,19,19},
    /* 28 ch */ {18,18,18,18,18,18,18,18,17,17,17,17,17,17,18,18,18,18,18,18,18,18,17,17,17,17,17,17},
    /* 30 ch */ {17,17,17,16,16,16,16,16,16,16,16,16,16,16,16,17,17,17,16,16,16,16,16,16,16,16,16,16,16,16},
    /* 32 ch */ {15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15}
};

int32_t getSamplePosition(uint8_t ch)
{
    int32_t pos;
    scope_t *sc;

    sc = &scope[ch];

    if (sc->latchFlag)
        return (-1); /* scope is latching at the moment, try again later */

    pos = sc->readPos;
    if ((sc->data != NULL) && (pos >= 0) && (pos < sc->len))
    {
        if (sc->sample16bit)
            pos *= 2;

        /* do this check again here - check if scope is latching at the moment */
        if (sc->latchFlag)
            return (-1);

        return (pos);
    }

    return (-1);
}

void stopAllScopes(void)
{
    uint8_t i;
    scope_t *sc;

    /* wait for scopes to finish updating */
    while (scopesUpdating);

    /* zero scope flags and stop scopes */
    for (i = 0; i < MAX_VOICES; ++i)
    {
        sc = &scope[i];

        sc->latchFlag = false;
        sc->data      = NULL;
    }

    /* wait for scope displaying to be done (safety) */
    while (scopesDisplaying);
}

void stopScope(uint8_t ch)
{
    scope_t *sc;

    /* wait for scope to finish updating */
    while (scopesUpdating);

    /* zero scope flags and stop scope */
    sc = &scope[ch];
    sc->latchFlag = false;
    sc->data      = NULL;

    /* wait for scope displaying to be done (safety) */
    while (scopesDisplaying);
}

/* toggle mute */
void setChannel(uint8_t nr, uint8_t on)
{
    stmTyp *ch;

    ch = &stm[nr];

    ch->stOff = !on;
    if (ch->stOff)
    {
        ch->effTyp   = 0;
        ch->eff      = 0;
        ch->realVol  = 0;
        ch->outVol   = 0;
        ch->oldVol   = 0;
        ch->finalVol = 0;
        ch->outPan   = 128;
        ch->oldPan   = 128;
        ch->finalPan = 128;
        ch->status   = IS_Vol;

        ch->envSustainActive = false; /* non-FT2 bug fix for stuck piano keys */
    }

    scope[nr].wasCleared = false;
}

static void drawScopeNumber(uint16_t scopeXOffs, uint16_t scopeYOffs, uint8_t channel, uint8_t outline)
{
    scopeYOffs++;
    channel++;

    if (outline)
    {
        if (channel < 10) /* one digit? */
        {
            charOutFastOutlined(scopeXOffs, scopeYOffs, PAL_MOUSEPT, '0' + channel);
        }
        else
        {
            charOutFastOutlined(scopeXOffs,     scopeYOffs, PAL_MOUSEPT, '0' + (channel / 10));
            charOutFastOutlined(scopeXOffs + 7, scopeYOffs, PAL_MOUSEPT, '0' + (channel % 10));
        }
    }
    else
    {
        if (channel < 10) /* one digit? */
        {
            charOutFast(scopeXOffs, scopeYOffs, PAL_MOUSEPT, '0' + channel);
        }
        else
        {
            charOutFast(scopeXOffs,     scopeYOffs, PAL_MOUSEPT, '0' + (channel / 10));
            charOutFast(scopeXOffs + 7, scopeYOffs, PAL_MOUSEPT, '0' + (channel % 10));
        }
    }
}

static void redrawScope(uint8_t ch)
{
    const uint8_t scopeMuteBMPWidths[16] =
    {
        162,111, 76, 56, 42, 35, 28, 24,
         21, 21, 17, 17, 12, 12,  9,  9
    };

    const uint8_t scopeMuteBMPHeights[16] =
    {
        27, 27, 26, 25, 25, 25, 24, 24,
        24, 24, 24, 24, 24, 24, 24, 24
    };

    const uint8_t *scopeMuteBMPPointers[16] =
    {
        scopeMuteBMP1, scopeMuteBMP2, scopeMuteBMP3, scopeMuteBMP4,
        scopeMuteBMP5, scopeMuteBMP6, scopeMuteBMP7, scopeMuteBMP8,
        scopeMuteBMP9, scopeMuteBMP9, scopeMuteBMP10,scopeMuteBMP10,
        scopeMuteBMP11,scopeMuteBMP11,scopeMuteBMP12,scopeMuteBMP12
    };

    int8_t chansPerRow;
    uint8_t i, chanLookup;
    int16_t scopeLen, muteGfxLen, muteGfxX;
    const uint16_t *scopeLens;
    uint16_t x, y;

    chansPerRow = song.antChn / 2;
    chanLookup  = chansPerRow - 1;
    scopeLens   = scopeLenTab[chanLookup];

    scopeLen = scopeLens[0];

    x = 2;
    y = 94;

    for (i = 0; i < song.antChn; ++i)
    {
        scopeLen = scopeLens[i];

        if (i == chansPerRow) /* did we reach end of row? */
        {
            /* yes, go one row down */
            x  = 2;
            y += 39;
        }

        if (i == ch)
            break;

        /* adjust position to next channel */
        x += (scopeLen + 3);
    }

    drawFramework(x, y, scopeLen + 2, 38, FRAMEWORK_TYPE2);

    /* draw mute graphics if channel is muted */
    if (editor.channelMute[i])
    {
        muteGfxLen = scopeMuteBMPWidths[chanLookup];
        muteGfxX   = x + ((scopeLen - muteGfxLen) / 2);

        blitFast(muteGfxX, y + 6, scopeMuteBMPPointers[chanLookup], muteGfxLen, scopeMuteBMPHeights[chanLookup]);

        /* draw channel number if needed */
        if (config.ptnChnNumbers)
            drawScopeNumber(x + 1, y + 1, i, true);
    }

    scope[ch].wasCleared = false;
}

void unmuteAllChansOnMusicLoad(void)
{
    uint8_t i;

    for (i = 0; i < MAX_VOICES; ++i)
    {
        editor.channelMute[i] = false;
        scope[i].wasCleared   = false;
    }
}

static void muteChannel(uint8_t ch)
{
    editor.channelMute[ch] = true;
    setChannel(ch, false);
    redrawScope(ch);
}

static void unmuteChannel(uint8_t ch)
{
    editor.channelMute[ch] = false;
    setChannel(ch, true);
    redrawScope(ch);
}

static void toggleChannelMute(uint8_t ch)
{
    editor.channelMute[ch] ^= 1;
    setChannel(ch, !editor.channelMute[ch]);
    redrawScope(ch);
}

int8_t testScopesMouseDown(void)
{
    int8_t chanToToggle, chansPerRow;
    uint8_t test, i;
    uint16_t x;
    const uint16_t *scopeLens;

    if (!editor.ui.scopesShown)
        return (false);

    if ((mouse.y >= 95) && (mouse.y <= 169) &&
        (mouse.x >=  3) && (mouse.x <= 288))
    {
        if ((mouse.y > 130) && (mouse.y < 134))
            return (true);

        chansPerRow = song.antChn / 2;
        scopeLens   = scopeLenTab[chansPerRow - 1];

        /* find out if we clicked inside a scope */
        x = 3;
        for (i = 0; i < chansPerRow; ++i)
        {
            if ((mouse.x >= x) && (mouse.x < (x + scopeLens[i])))
                break;

            x += (scopeLens[i] + 3);
        }

        if (i == chansPerRow)
            return (true); /* scope framework was clicked instead */

        chanToToggle = i;
        if (mouse.y >= 134) /* second row of scopes? */
            chanToToggle += chansPerRow; /* yes, increase lookup offset */

        MY_ASSERT(chanToToggle < song.antChn)

        if (mouse.leftButtonPressed && mouse.rightButtonPressed)
        {
            test = false;
            for (i = 0; i < song.antChn; ++i)
            {
                if ((i != chanToToggle) && editor.channelMute[i])
                    test = true;
            }

            if (test)
            {
                for (i = 0; i < song.antChn; ++i)
                    unmuteChannel(i);
            }
            else
            {
                for (i = 0; i < song.antChn; ++i)
                {
                    if (i == chanToToggle)
                        unmuteChannel(i);
                    else
                        muteChannel(i);
                }
            }
        }
        else if (mouse.leftButtonPressed && !mouse.rightButtonPressed)
        {
            toggleChannelMute(chanToToggle);
        }
        else
        {
            /* toggle channel recording */
            config.multiRecChn[chanToToggle] ^= 1;
            redrawScope(chanToToggle);
        }

        return (true);
    }

    return (false);
}

void setScopeRate(uint8_t channel, uint32_t rate)
{
    scope_t *s;

    s = &scope[channel];

    if (rate == 0)
    {
        s->fDelta     = 0.0f;
        s->dDrawDelta = 0.0;
    }
    else
    {
        s->fDelta     = rate / (float)(VBLANK_HZ);
        s->dDrawDelta = rate / 4096.0; /* FT2 is using 4096 here, which is 2^n closest to C-3's Hz (4181) */
    }
}

void setScopeVolume(uint8_t channel, uint16_t vol)
{
    scope[channel].volume = (int8_t)((vol * SCOPE_DATA_HEIGHT) / 256);
}

void triggerScope(uint8_t i, int8_t *pek, int32_t len, int32_t repS, int32_t repL, uint8_t typ, int32_t playOffset)
{
    scopeState_t *newState;

    newState = &scopeNewState[i];

    newState->pek  = pek;
    newState->len  = len;
    newState->repS = repS;
    newState->repL = repL;
    newState->typ  = typ;
    newState->playOffset = playOffset;

    scope[i].latchFlag = true;
}

static void triggerInternalScope(uint8_t channel)
{
    int8_t *sampleData;
    uint8_t loopFlag, sampleIs16Bit;
    int32_t sampleLength, sampleLoopBegin, sampleLoopLength, sampleLoopEnd, position;
    scope_t *sc;
    scopeState_t *s;

    sc = &scope[channel];
    s  = &scopeNewState[channel];

    sampleData       = s->pek;
    sampleLength     = s->len;
    sampleLoopBegin  = s->repS;
    sampleLoopLength = s->repL;
    sampleLoopEnd    = s->repS + s->repL;
    loopFlag         = s->typ & 3;
    sampleIs16Bit    = s->typ & 16;
    position         = s->playOffset;

    if ((sampleData == NULL) || (sampleLength < 1))
    {
        sc->data = NULL; /* shut down scope */
        return;
    }

    if (sampleIs16Bit)
    {
        if (sampleLoopLength < 2)
            loopFlag = false;

        MY_ASSERT(!(sampleLoopBegin  & 1))
        MY_ASSERT(!(sampleLength     & 1))
        MY_ASSERT(!(sampleLoopLength & 1))
        MY_ASSERT(!(sampleLoopEnd    & 1))

        sampleLoopBegin  /= 2;
        sampleLength     /= 2;
        sampleLoopLength /= 2;
        sampleLoopEnd    /= 2;
    }
    else
    {
        if (sampleLoopLength < 1)
            loopFlag = false;
    }

    /* test if 9xx position overflows */
    if (loopFlag)
    {
        if (position >= sampleLoopEnd)
        {
            sc->data = NULL; /* shut down scope */
            return;
        }
    }
    else
    {
        if (position >= sampleLength)
        {
            sc->data = NULL; /* shut down scope */
            return;
        }
    }

    sc->direction   = 1; /* forward (-1 = backwards) */
    sc->len         = sampleLength;
    sc->loop        = loopFlag;
    sc->loopStart   = sampleLoopBegin;
    sc->loopEnd     = sampleLoopEnd;
    sc->loopLen     = sampleLoopLength;
    sc->readPos     = position;
    sc->sample16bit = sampleIs16Bit;
    sc->fFrac       = 0.0f;

    sc->data = s->pek;
}

void updateScopes(void)
{
    uint8_t i;
    int32_t intFrac, readPos, overflow;
    scope_t *sc;

    scopesUpdating = true;
    for (i = 0; i < song.antChn; ++i)
    {
        sc = &scope[i];

        if (sc->latchFlag)
        {
            triggerInternalScope(i);
            sc->latchFlag = false; /* we can now read sampling position from scopes in other threads */

            continue; /* don't increase sampling position on trigger frame */
        }

        /* scope position update */

        /* from now on, only continue if the scope is active */
        if (sc->data == NULL)
            continue;

        sc->fFrac += sc->fDelta;
        if (sc->fFrac >= 1.0f)
        {
            readPos = sc->readPos; /* this one has to be copied and updated when done */

            intFrac = (int32_t)(sc->fFrac); /* truncate (uses SSE) */
            sc->fFrac -= intFrac;
            readPos += (intFrac * sc->direction);

            if (sc->direction == -1) /* definitely bidi loop */
            {
                if (readPos < sc->loopStart)
                {
                    sc->direction = 1;

                    if (sc->loopLen < 2)
                        readPos = sc->loopStart;
                    else
                        readPos = sc->loopStart + ((sc->loopStart - readPos - 1) % sc->loopLen);
                }
            }
            else if (sc->loop == 0) /* no loop */
            {
                if (readPos >= sc->len)
                    sc->data = NULL;
            }
            else if (readPos >= sc->loopEnd) /* forward or bidi loop */
            {
                overflow = 0;
                if (sc->loopLen >= 2)
                    overflow = (readPos - sc->loopEnd) % sc->loopLen;

                if (sc->loop == 2) /* bidi loop */
                {
                    sc->direction = -1;
                    readPos = (sc->loopEnd - 1) - overflow;
                }
                else /* forward loop */
                {
                    readPos = sc->loopStart + overflow;
                }
            }

            MY_ASSERT((sc->data == NULL) || ((readPos >= 0) && (readPos < sc->len)))

            sc->readPos = readPos;
        }
    }
    scopesUpdating = false;
}

/* this routine should ONLY be used inside the scoop drawing loop! */
static int8_t getScopeSample(scope_t *sc, int32_t samplePos)
{
    const void *data;
    int8_t sample;
    const int8_t *data8;
    const int16_t *data16;
    int32_t tmp32;

    data = sc->data;
    if (data == NULL)
        return (0);

    if (sc->loop == 0) /* no loop */
    {
        if (samplePos >= sc->len)
            return (0);
    }
    else if (sc->drawDirection == -1) /* definitely bidi loop */
    {
        if (samplePos < sc->loopStart)
        {
            sc->drawDirection = 1;

            if (sc->loopLen < 2)
                samplePos = sc->loopStart;
            else
                samplePos = sc->loopStart + ((sc->loopStart - samplePos - 1) % sc->loopLen);
        }
    }
    else if (samplePos >= sc->loopEnd) /* forward or bidi loop */
    {
        tmp32 = 0;
        if (sc->loopLen >= 2)
            tmp32 += ((samplePos - sc->loopEnd) % sc->loopLen);

        if (sc->loop == 2) /* bidi loop */
        {
            sc->drawDirection = -1;
            samplePos = (sc->loopEnd - 1) - tmp32;
        }
        else /* forward loop */
        {
            samplePos = sc->loopStart + tmp32;
        }
    }

    MY_ASSERT((samplePos >= 0) && (samplePos < sc->len))

    if (sc->sample16bit)
    {
        /* 16-bit sample */
        data16 = (const int16_t *)(data);
        sample = (int8_t)((data16[samplePos] * sc->volume) >> 16);
    }
    else
    {
        /* 8-bit sample */
        data8  = (const int8_t *)(data);
        sample = (int8_t)((data8[samplePos] * sc->volume) >> 8);
    }

    return (sample);
}

static void scopeLine(int16_t x1, int16_t y1, int16_t y2)
{
    int16_t d, sy, dy;
    uint16_t ay;
    int32_t pitch;
    uint32_t pixVal, *dst32;

    dy = y2 - y1;
    ay = ABS(dy);
    sy = SGN(dy);

    pixVal = video.palette[PAL_PATTEXT];
    pitch  = sy * SCREEN_W;

    dst32  = &video.frameBuffer[(y1 * SCREEN_W) + x1];
    *dst32 = pixVal;

    if (ay <= 1)
    {
        if (ay != 0)
            dst32 += pitch;

        *++dst32 = pixVal;
    }
    else
    {
        d = 2 - ay;

        ay *= 2;
        while (y1 != y2)
        {
            if (d >= 0)
            {
                d -= ay;
                dst32++;
            }

            y1 += sy;
            d  += 2;

             dst32 += pitch;
            *dst32  = pixVal;
        }
    }
}

/* ---------------------------------------------- */
#define SCOPE_NEXT_SAMPLE() \
dScopeDrawPos += (sc->dDrawDelta * sc->drawDirection); \
scopeDrawPos = (int32_t)(dScopeDrawPos); /* truncate */
/* ---------------------------------------------- */

void drawScopes(void)
{
    uint8_t i;
    int16_t y1, y2, sample, scopeLineY;
    const uint16_t *scopeLens;
    uint16_t x16, scopeXOffs, scopeYOffs, scopeDrawLen;
    int32_t scopeDrawPos;
    uint32_t x, len, scopePixelColor;
    double dScopeDrawPos;
    scope_t cachedScope, *sc;

    scopesDisplaying = true;

    scopeLens  = scopeLenTab[(song.antChn / 2) - 1];
    scopeXOffs = 3;
    scopeYOffs = 95;
    scopeLineY = 112;

    for (i = 0; i < song.antChn; ++i)
    {
        /* check if we are at the bottom row */
        if (i == (song.antChn / 2))
        {
            scopeXOffs = 3;
            scopeYOffs = 134;
            scopeLineY = 151;
        }

        scopeDrawLen = scopeLens[i];

        if (!editor.channelMute[i])
        {
            sc = &scope[i];

            sc->drawDirection = sc->direction;

            /* cache scope channel to lower thread race condition issues */
            memcpy(&cachedScope, sc, sizeof (scope_t));
            sc = &cachedScope;

            if ((sc->data != NULL) && (sc->volume > 0) && !audio.locked)
            {
                /* scope is active */

                scope[i].wasCleared = false;

                /* clear scope background */
                clearRect(scopeXOffs, scopeYOffs, scopeDrawLen, SCOPE_DATA_HEIGHT);

                scopeDrawPos  = sc->readPos;
                dScopeDrawPos = scopeDrawPos;

                /* draw current scope */
                if (config.specialFlags & LINED_SCOPES)
                {
                    /* LINE SCOPE */

                    y1 = scopeLineY - getScopeSample(sc, scopeDrawPos);
                    SCOPE_NEXT_SAMPLE();

                    x16 = scopeXOffs;
                    len = scopeXOffs + (scopeDrawLen - 1);

                    for (; x16 < len; ++x16)
                    {
                        y2 = scopeLineY - getScopeSample(sc, scopeDrawPos);
                        scopeLine(x16, y1, y2);
                        y1 = y2;

                        SCOPE_NEXT_SAMPLE();
                    }
                }
                else
                {
                    /* PIXEL SCOPE */

                    scopePixelColor = video.palette[PAL_PATTEXT];

                    x   = scopeXOffs;
                    len = scopeXOffs + scopeDrawLen;

                    for (; x < len; ++x)
                    {
                        sample = getScopeSample(sc, scopeDrawPos);
                        video.frameBuffer[((scopeLineY - sample) * SCREEN_W) + x] = scopePixelColor;

                        SCOPE_NEXT_SAMPLE();
                    }
                }
            }
            else
            {
                /* scope is inactive */

                sc = &scope[i];
                if (!sc->wasCleared)
                {
                    /* clear scope background */
                    clearRect(scopeXOffs, scopeYOffs, scopeDrawLen, SCOPE_DATA_HEIGHT);

                    /* draw empty line */
                    hLine(scopeXOffs, scopeLineY, scopeDrawLen, PAL_PATTEXT);

                    sc->wasCleared = true;
                }
            }

            /* draw channel numbering (if enabled) */
            if (config.ptnChnNumbers)
                drawScopeNumber(scopeXOffs, scopeYOffs, i, false);

            /* draw rec. symbol (if enabled) */
            if (config.multiRecChn[i])
                blit(scopeXOffs, scopeYOffs + 31, scopeRecBMP, 13, 4);
        }

        /* align x to next scope */
        scopeXOffs += (scopeDrawLen + 3);
    }

    scopesDisplaying = false;
}

void drawScopeFramework(void)
{
    uint8_t i;

    drawFramework(0, 92, 291, 81, FRAMEWORK_TYPE1);
    for (i = 0; i < song.antChn; ++i)
        redrawScope(i);
}

void handleScopesFromChQueue(chSyncData_t *chSyncData, uint8_t *scopeUpdateStatus)
{
    uint8_t i, status;
    channel_t *ch;
    sampleTyp *s;

    for (i = 0; i < song.antChn; ++i)
    {
        ch = &chSyncData->channels[i];
        status = scopeUpdateStatus[i];

        if (status & IS_Vol)    setScopeVolume(i, ch->finalVol);
        if (status & IS_Period) setScopeRate(i, getFrequenceValueScope(ch->finalPeriod));

        if (status & IS_NyTon) /* trigger sample */
        {
            s = ch->smpPtr;
            if (s != NULL)
            {
                if (ch->instrNr == (MAX_INST + 1))
                {
                    /* sample editor ("Wave, "Range", "Display") */
                    triggerScope(i, s->pek, s->len, s->repS, s->repL, s->typ, ch->smpStartPos + editor.samplePlayOffset);
                    editor.samplePlayOffset = 0;
                }
                else
                {
                    /* normal note jamming */
                    triggerScope(i, s->pek, s->len, s->repS, s->repL, s->typ, ch->smpStartPos);
                }

                /* set curr. channel for getting sampling position line (sample editor) */
                if ((ch->instrNr == (MAX_INST + 1)) || ((editor.curInstr == ch->instrNr) && (editor.curSmp == ch->sampleNr)))
                    editor.curSmpChannel = i;
            }
        }
    }
}

int32_t SDLCALL scopeThreadFunc(void *ptr)
{
    int32_t time32;
    uint32_t diff32;
    uint64_t time64;
    double dTime;

#ifdef _WIN32
    /* this is needed for the scopes to stutter slightly less (confirmed) */
    SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH);
#endif

    /* set next frame time */
    timeNext64 = SDL_GetPerformanceCounter() + video.vblankTimeLen;

    while (editor.programRunning)
    {
        editor.scopeThreadMutex = true;
        updateScopes();
        editor.scopeThreadMutex = false;

        time64 = SDL_GetPerformanceCounter();
        if (time64 < timeNext64)
        {
            MY_ASSERT((timeNext64 - time64) <= 0xFFFFFFFFULL)
            diff32 = (uint32_t)(timeNext64 - time64);

#ifdef _WIN32
            /* convert to milliseconds */
            dTime = diff32 * editor.dPerfFreqMulMilli;
            double2int32_round(time32, dTime);

            /* delay until we have reached next tick */
            if (time32 > 0)
                SDL_Delay(time32);
#else
            /* convert to microseconds */
            dTime = diff32 * editor.dPerfFreqMulMicro;
            double2int32_round(time32, dTime);

            /* delay until we have reached next tick */
            if (time32 > 0)
                usleep(time32);
#endif
        }

        /* update next tick time */
        timeNext64 += video.vblankTimeLen;
    }

    (void)(ptr); /* make compiler happy */

    return (true);
}

uint8_t initScopes(void)
{
    scopeThread = SDL_CreateThread(scopeThreadFunc, "FT2 Clone Visuals Thread", NULL);
    if (scopeThread == NULL)
        return (false);

    /* don't let thread wait for this thread, let it clean up on its own when done */
    SDL_DetachThread(scopeThread);

    return (true);
}
